#!/usr/bin/env python3
import argparse
import json
import logging
import os
import signal
import time
from functools import partial
from typing import Optional

import coloredlogs
from Kathara.exceptions import MachineCollisionDomainError
from Kathara.manager.Kathara import Kathara
from Kathara.model.Lab import Lab
from Kathara.parser.netkit.LabParser import LabParser
from Kathara.setting.Setting import Setting
from tqdm import tqdm

from .TestCollector import TestCollector
from .checks.BridgeCheck import BridgeCheck
from .checks.CheckResult import CheckResult
from .checks.CollisionDomainCheck import CollisionDomainCheck
from .checks.CustomCommandCheck import CustomCommandCheck
from .checks.DaemonCheck import DaemonCheck
from .checks.DeviceExistenceCheck import DeviceExistenceCheck
from .checks.IPv6EnabledCheck import IPv6EnabledCheck
from .checks.InterfaceIPCheck import InterfaceIPCheck
from .checks.KernelRouteCheck import KernelRouteCheck
from .checks.ReachabilityCheck import ReachabilityCheck
from .checks.StartupExistenceCheck import StartupExistenceCheck
from .checks.SysctlCheck import SysctlCheck
from .checks.applications.dns.DNSAuthorityCheck import DNSAuthorityCheck
from .checks.applications.dns.DNSRecordCheck import DNSRecordCheck
from .checks.applications.dns.LocalNSCheck import LocalNSCheck
from .checks.protocols.AnnouncedNetworkCheck import AnnouncedNetworkCheck
from .checks.protocols.ProtocolRedistributionCheck import ProtocolRedistributionCheck
from .checks.protocols.bgp.BGPPeeringCheck import BGPPeeringCheck
from .checks.protocols.evpn.AnnouncedVNICheck import AnnouncedVNICheck
from .checks.protocols.evpn.EVPNSessionCheck import EVPNSessionCheck
from .checks.protocols.evpn.VTEPCheck import VTEPCheck
from .excel_utils import write_final_results_to_excel, write_result_to_excel
from .utils import reverse_dictionary

VERSION = "0.1.6"
CURRENT_LAB: Optional[Lab] = None


def handler(signum, frame, live=False):
    logger = logging.getLogger("kathara-lab-checker")
    if CURRENT_LAB and not live:
        logger.warning(f"\nCtrl-C was pressed. Undeploying current lab in: {CURRENT_LAB.fs_path()}")
        Kathara.get_instance().undeploy_lab(lab=CURRENT_LAB)
    exit(1)


def run_on_single_network_scenario(
        lab_path: str,
        configuration: dict,
        lab_template: Lab,
        no_cache: bool = False,
        live: bool = False,
        keep_open: bool = False,
        skip_report: bool = False,
):
    global CURRENT_LAB
    logger = logging.getLogger("kathara-lab-checker")

    manager = Kathara.get_instance()
    test_collector = TestCollector()

    lab_path = os.path.abspath(lab_path)
    lab_name = os.path.basename(lab_path)

    if not os.path.isdir(lab_path):
        logger.warning(f"{lab_path} is not a lab directory.")
        return

    test_results_path = os.path.join(lab_path, f"{lab_name}_result.xlsx")
    if os.path.exists(test_results_path) and not no_cache:
        logger.warning("Network scenario already processed, skipping...")
        return

    logger.info(f"##################### {lab_name} #####################")
    logger.info(f"Parsing network scenario in: {lab_path}")
    logger.info(f"Network scenario name: {lab_name}")
    try:
        lab = LabParser().parse(lab_path)
        CURRENT_LAB = lab
    except IOError as e:
        logger.warning(f"{str(e)} Skipping directory")
        check_results = [CheckResult("The lab.conf cannot be parsed", False, str(e))]
        test_collector.add_check_results(lab_name, check_results)
        return test_collector
    except MachineCollisionDomainError as e:
        logger.warning(f"{str(e)} Skipping directory")
        check_results = [CheckResult("The lab.conf cannot be parsed", False, str(e))]
        test_collector.add_check_results(lab_name, check_results)
        return test_collector

    if not live:
        logger.info(f"Undeploying network scenario in case it was running...")
        manager.undeploy_lab(lab=lab)
        logger.info(f"Deploying network scenario...")
        manager.deploy_lab(lab=lab)

        logger.info(f"Waiting convergence...")
        time.sleep(configuration["convergence_time"])
    else:
        machines = manager.get_machines_api_objects(lab=lab)
        if not machines:
            logger.warning("No devices running in the network scenario. Test aborted.")
            return

    logger.info(f"Starting tests")

    logger.info(f"Verifying lab structure using lab.conf template in: {configuration['structure']}")

    logger.info("Checking that all devices exist...")
    check_results = DeviceExistenceCheck().run(list(lab_template.machines.keys()), lab)
    test_collector.add_check_results(lab_name, check_results)

    logger.info("Checking collision domains...")
    check_results = CollisionDomainCheck().run(list(lab_template.machines.values()), lab)
    test_collector.add_check_results(lab_name, check_results)

    if "requiring_startup" in configuration["test"]:
        logger.info("Checking that all required startup files exist...")
        check_results = StartupExistenceCheck().run(configuration["test"]["requiring_startup"], lab)
        test_collector.add_check_results(lab_name, check_results)

    if "ipv6_enabled" in configuration["test"]:
        logger.info(f"Checking that IPv6 is enabled on devices: {configuration['test']['ipv6_enabled']}")
        check_results = IPv6EnabledCheck().run(configuration["test"]["ipv6_enabled"], lab)
        test_collector.add_check_results(lab_name, check_results)

    if "sysctls" in configuration["test"]:
        logger.info(f"Checking sysctl configurations on devices...")
        check_results = SysctlCheck().run(configuration["test"]["sysctls"], lab)
        test_collector.add_check_results(lab_name, check_results)

    if "ip_mapping" in configuration["test"]:
        logger.info("Verifying the IP addresses assigned to devices...")
        check_results = InterfaceIPCheck().run(configuration["test"]["ip_mapping"], lab)
        test_collector.add_check_results(lab_name, check_results)

    if "bridges" in configuration["test"]:
        logger.info("Verifying the bridges inside devices...")
        check_results = BridgeCheck().run(configuration["test"]["bridges"], lab)
        test_collector.add_check_results(lab_name, check_results)

    if "reachability" in configuration["test"]:
        logger.info(f"Starting reachability test...")
        check_results = ReachabilityCheck().run(configuration["test"]["reachability"], lab)
        test_collector.add_check_results(lab_name, check_results)

    if "daemons" in configuration["test"]:
        logger.info(f"Checking if daemons are running...")
        check_results = DaemonCheck().run(configuration["test"]["daemons"], lab)
        test_collector.add_check_results(lab_name, check_results)

    if "protocols" in configuration["test"]:
        logger.info("Checking routing daemons configurations...")
        for daemon_name, daemon_test in configuration["test"]["protocols"].items():
            if daemon_name == "bgpd":
                logger.info(f"Check BGP peerings configurations...")
                check_results = BGPPeeringCheck().run(daemon_test["peerings"], lab)
                test_collector.add_check_results(lab_name, check_results)

                if "networks" in daemon_test:
                    logger.info(f"Checking BGP announces...")
                    check_results = AnnouncedNetworkCheck().run(daemon_name, daemon_test["networks"], lab)
                    test_collector.add_check_results(lab_name, check_results)

                if "evpn" in daemon_test:
                    logger.info(f"Checking EVPN configurations...")
                    evpn_test = daemon_test["evpn"]
                    for test in evpn_test:
                        if "evpn_sessions" in test:
                            logger.info(f"Checking EVPN session configuration...")
                            check_results = EVPNSessionCheck().run(evpn_test["evpn_sessions"], lab)
                            test_collector.add_check_results(lab_name, check_results)

                        if "vtep_devices" in test:
                            logger.info(f"Checking VTEP devices configuration...")
                            check_results = VTEPCheck().run(evpn_test["vtep_devices"], lab)
                            test_collector.add_check_results(lab_name, check_results)

                            logger.info(f"Checking BGP VNIs configurations...")
                            check_results = AnnouncedVNICheck().run(
                                evpn_test["vtep_devices"], evpn_test["evpn_sessions"], lab
                            )
                            test_collector.add_check_results(lab_name, check_results)

            if "injections" in daemon_test:
                logger.info(f"Checking {daemon_name} protocols redistributions...")
                check_results = ProtocolRedistributionCheck().run(daemon_name, daemon_test["injections"], lab)
                test_collector.add_check_results(lab_name, check_results)

    if "kernel_routes" in configuration["test"]:
        logger.info(f"Checking Routing Tables...")
        check_results = KernelRouteCheck().run(configuration["test"]["kernel_routes"], lab)
        test_collector.add_check_results(lab_name, check_results)

    if "applications" in configuration["test"]:
        for application_name, application in configuration["test"]["applications"].items():
            if application_name == "dns":
                if "authoritative" in application:
                    logger.info("Checking DNS configurations...")
                    check_results = DNSAuthorityCheck().run(
                        application["authoritative"],
                        list(application["local_ns"].keys()),
                        configuration["test"]["ip_mapping"],
                        lab,
                    )
                    test_collector.add_check_results(lab_name, check_results)

                if "local_ns" in application:
                    logger.info("Checking local name servers configurations...")
                    check_results = LocalNSCheck().run(application["local_ns"], lab)
                    test_collector.add_check_results(lab_name, check_results)

                if "records" in application:
                    logger.info(f"Starting test for DNS records...")
                    check_results = DNSRecordCheck().run(
                        application["records"], reverse_dictionary(application["local_ns"]).keys(), lab
                    )
                    test_collector.add_check_results(lab_name, check_results)

    if "custom_commands" in configuration["test"]:
        logger.info("Checking custom commands output...")
        check_results = CustomCommandCheck().run(configuration["test"]["custom_commands"], lab)
        test_collector.add_check_results(lab_name, check_results)

    if not live and not keep_open:
        logger.info("Undeploying Network Scenario")
        manager.undeploy_lab(lab=lab)

    total_tests = len(test_collector.tests[lab_name])
    test_results = list(map(lambda x: x.passed, test_collector.tests[lab_name]))
    logger.info(f"Total Tests: {total_tests}")
    logger.info(f"Passed Tests: {test_results.count(True)}/{total_tests}")

    if not skip_report:
        logger.info(f"Writing test report for {lab_name} in: {lab_path}...")
        write_result_to_excel(test_collector.tests[lab_name], lab_path)

    return test_collector


def run_on_multiple_network_scenarios(
        labs_path: str,
        configuration: dict,
        lab_template: Lab,
        no_cache: bool = False,
        live: bool = False,
        keep_open: bool = False,
        skip_report: bool = False,
):
    logger = logging.getLogger("kathara-lab-checker")
    labs_path = os.path.abspath(labs_path)

    logger.info(f"Parsing network scenarios in: {labs_path}")

    test_collector = TestCollector()
    for lab_name in tqdm(
            sorted(
                list(
                    filter(
                        lambda x: os.path.isdir(os.path.join(labs_path, x)) and x != ".DS_Store",
                        os.listdir(labs_path),
                    )
                )
            )
    ):
        test_results = run_on_single_network_scenario(
            os.path.join(labs_path, lab_name), configuration, lab_template, no_cache, live, keep_open, skip_report
        )

        if test_results:
            test_collector.add_check_results(lab_name, test_results.tests[lab_name])

    if test_collector.tests and not skip_report:
        logger.info(f"Writing All Test Results into: {labs_path}")
        write_final_results_to_excel(test_collector, labs_path)


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="A tool for automatically check Kathará network scenarios",
        prog="kathara_lab_checker",
        add_help=True
    )

    parser.add_argument(
        "--config",
        "-c",
        required=True,
        help="The path to the configuration file for the tests",
    )

    parser.add_argument("-v", "--version", action="version", version=f"kathara-lab-checker {VERSION}")

    parser.add_argument(
        "--no-cache",
        required=False,
        action="store_true",
        default=False,
        help="Re-process all the tests",
    )

    parser.add_argument(
        "--live",
        required=False,
        action="store_true",
        default=False,
        help="Do not deploy/undeploy the network scenarios",
    )

    parser.add_argument(
        "--keep-open",
        required=False,
        action="store_true",
        default=False,
        help="Do not undeploy the network scenarios",
    )

    group = parser.add_mutually_exclusive_group(required=True)

    group.add_argument(
        "--lab",
        required=False,
        help="The path to the network scenario to check",
    )

    group.add_argument(
        "--labs",
        required=False,
        help="The path to a directory containing multiple network scenarios to check with the same configuration",
    )

    parser.add_argument(
        "--skip-report",
        required=False,
        action="store_true",
        default=False,
        help="Skip the generation of the report",
    )

    return parser.parse_args()


def main():
    args = parse_arguments()

    signal.signal(signal.SIGINT, partial(handler, live=args.live))

    logger = logging.getLogger("kathara-lab-checker")
    # logger.addHandler(TqdmLoggingHandler())

    coloredlogs.install(fmt="%(message)s", level="INFO", logger=logger)

    logger.propagate = False

    logger.info("Parsing test configuration...")
    with open(args.config, "r") as json_conf:
        conf = json.load(json_conf)

    Setting.get_instance().load_from_dict({"image": conf["default_image"]})

    logger.info(f"Parsing network scenarios template in: {conf['structure']}")
    if not os.path.exists(conf["structure"]):
        logger.error(f"The structure file {conf['structure']} does not exist")
        exit(1)

    template_lab = LabParser().parse(
        os.path.dirname(conf["structure"]),
        conf_name=os.path.basename(conf["structure"]),
    )

    if args.lab:
        run_on_single_network_scenario(
            args.lab, conf, template_lab, args.no_cache, args.live, args.keep_open, args.skip_report
        )
    elif args.labs:
        run_on_multiple_network_scenarios(
            args.labs, conf, template_lab, args.no_cache, args.live, args.keep_open, args.skip_report
        )


if __name__ == "__main__":
    main()



from Kathara.exceptions import LinkNotFoundError
from Kathara.model.Lab import Lab
from Kathara.model.Machine import Machine

from .AbstractCheck import AbstractCheck
from .CheckResult import CheckResult


class CollisionDomainCheck(AbstractCheck):
    def check(self, machine_t: Machine, lab: Lab) -> list[CheckResult]:

        results = []
        try:
            machine = lab.get_machine(machine_t.name)
            for iface_num, interface in machine.interfaces.items():
                self.description = f"Checking the collision domain attached to interface `eth{iface_num}` of `{machine_t.name}`"
                interface_t = machine_t.interfaces[iface_num]
                if interface_t.link.name != interface.link.name:
                    reason = (
                        f"Interface `{iface_num}` of device {machine_t.name} is connected to collision domain "
                        f"`{interface.link.name}` instead of `{interface_t.link.name}`"
                    )
                    results.append(CheckResult(self.description, False, reason))
                else:
                    results.append(CheckResult(self.description, True, "OK"))


        except LinkNotFoundError as e:
            results.append(CheckResult(self.description, False, str(e)))
        return results

    def run(self, template_machines: list[Machine], lab: Lab) -> list[CheckResult]:
        results = []
        for machine_t in template_machines:
            check_result = self.check(machine_t, lab)
            results.extend(check_result)
        return results



import json

from Kathara.exceptions import MachineNotRunningError
from Kathara.manager.Kathara import Kathara
from Kathara.model.Lab import Lab

from kathara_lab_checker.utils import get_output
from .AbstractCheck import AbstractCheck
from .CheckResult import CheckResult


def filter_by_interface_type(interface_type: str, interfaces: list[dict]):
    return list(
        filter(lambda x: "linkinfo" in x and x["linkinfo"]["info_kind"] == interface_type, interfaces)
    )


def get_interface_by_name(interface_name: str, interfaces: list[dict]):
    return list(filter(lambda x: x["ifname"] == interface_name, interfaces)).pop()


def get_inteface_by_vni(interface_vni: str, interfaces: list[dict]):
    return list(
        filter(
            lambda x: "linkinfo" in x
                      and "id" in x["linkinfo"]["info_data"]
                      and x["linkinfo"]["info_data"]["id"] == int(interface_vni),
            interfaces,
        )
    ).pop()


class BridgeCheck(AbstractCheck):
    def check_bridge_interfaces(
            self, device_name: str, expected_interfaces: list[str], actual_interfaces: list[dict]
    ) -> (CheckResult, set[str]):
        self.description = (
            f"Checking that interfaces {expected_interfaces} "
            f"are attached to the same bridge on `{device_name}`"
        )
        interface_masters = {}
        interfaces_not_found = []
        interfaces_without_bridge = []
        for interface_name in expected_interfaces:
            try:
                actual_iface_info = get_interface_by_name(interface_name, actual_interfaces)
                if "master" in actual_iface_info:
                    interface_masters[interface_name] = actual_iface_info["master"]
                else:
                    interfaces_without_bridge.append(interface_name)
            except IndexError:
                interfaces_not_found.append(interface_name)

        if interfaces_not_found:
            return (
                CheckResult(
                    self.description,
                    False,
                    f"Interfaces `{interfaces_not_found}` are not found on {device_name}",
                ),
                None,
            )
        if interfaces_without_bridge:
            return (
                CheckResult(
                    self.description,
                    False,
                    f"Interfaces `{interfaces_without_bridge}` are not connected to any bridge on {device_name}",
                ),
                None,
            )
        masters = set(interface_masters.values())
        masters_num = len(masters)
        if masters_num == 1:
            master = list(masters)[0]
            if get_interface_by_name(master, actual_interfaces)["linkinfo"]["info_kind"]:
                return CheckResult(self.description, True, "OK"), masters
        elif masters_num == 0:
            return CheckResult(self.description, False, "No interfaces attached to the bridge"), masters
        elif masters_num > 1:
            reason = "Interfaces are not attached to the same bridge.\n"
            for interface_name, interface_master in interface_masters.items():
                master_type = get_interface_by_name(interface_master, actual_interfaces)["linkinfo"][
                    "info_kind"
                ]
                reason += f"`{interface_name}` to `{interface_master}` (type: {master_type})\n"
            return CheckResult(self.description, False, reason), masters

    def check_vlan_filtering(self, device_name: str, bridge_info: dict) -> CheckResult:
        self.description = (
            f"Checking if VLAN filtering is enabled on `{bridge_info['ifname']}` of `{device_name}`"
        )
        if (
                "vlan_filtering" in bridge_info["linkinfo"]["info_data"]
                and bridge_info["linkinfo"]["info_data"]["vlan_filtering"] == 1
        ):
            return CheckResult(self.description, True, "OK")
        else:
            return CheckResult(
                self.description,
                True,
                f"VLAN filtering not enabled on `{bridge_info['ifname']}` of `{device_name}`",
            )

    def check_vlan_tags(
            self,
            device_name: str,
            interface_name: str,
            interface_configuration: dict,
            actual_interface_vlan: dict,
    ):
        self.description = (
            f"Checking that vlans `{interface_configuration['vlan_tags']}` "
            f"are configured on `{interface_name}` of `{device_name}`"
        )
        expected_vlans = set(interface_configuration["vlan_tags"])
        actual_vlans = set(map(lambda x: x["vlan"], actual_interface_vlan["vlans"]))
        actual_vlans.remove(1)

        if expected_vlans == actual_vlans:
            return CheckResult(self.description, True, "OK")
        else:
            not_configured = expected_vlans.difference(actual_vlans)
            reason = f"Vlans `{not_configured}` are not configured on `{interface_name}` of `{device_name}`"
            return CheckResult(self.description, False, reason)

    def check_vxlan_pvid(
            self, device_name: str, vni: str, pvid: str, actual_interfaces: list[dict], vlans_info: list[dict]
    ):
        self.description = f"Checking that `{device_name}` manages VNI `{vni}` with PVID `{pvid}`"

        try:
            interface_name = get_inteface_by_vni(vni, actual_interfaces)["ifname"]
        except IndexError:
            return CheckResult(self.description, False, f"VNI {vni} not configured on `{device_name}`")

        for vlan in vlans_info:
            if vlan["ifname"] == interface_name:
                actual_pvid = set(
                    map(lambda x: x["vlan"], filter(lambda x: "PVID" in x["flags"], vlan["vlans"]))
                )
                if pvid:
                    actual_pvid = actual_pvid.pop()
                    if actual_pvid == pvid:
                        return CheckResult(self.description, True, "OK")
                    else:
                        return CheckResult(
                            self.description,
                            False,
                            f"VNI `{vni}` found with pvid `{actual_pvid}` (instead of {pvid})",
                        )
        return CheckResult(self.description, False, f"VNI `{vni}` not found on `{device_name}`")

    def check_vlan_pvid(
            self, device_name: str, interface_name: str, interface_pvid: str, actual_interface_vlan: dict
    ):
        self.description = f"Checking that `{interface_name}` of `{device_name}` has pvid {interface_pvid}"
        pvid = set(
            map(lambda x: x["vlan"], filter(lambda x: "PVID" in x["flags"], actual_interface_vlan["vlans"]))
        )
        if pvid:
            actual_pvid = pvid.pop()
            if interface_pvid == actual_pvid:
                return CheckResult(self.description, True, "OK")
            else:
                reason = (
                    f"`{interface_name}` of `{device_name}` has pvid `{actual_pvid}` "
                    f"(instead of `{interface_pvid}`)"
                )
                return CheckResult(self.description, False, reason)
        else:
            reason = f"No pvid configured on `{interface_name}` of `{device_name}`"
            return CheckResult(self.description, False, reason)

    def run(self, devices_to_bridge_configuration: dict[str, list[dict]], lab: Lab) -> list[CheckResult]:
        results = []
        for device_name, bridges_configuration in devices_to_bridge_configuration.items():
            self.logger.info(f"Checking bridges configuration on `{device_name}`...")
            try:
                ip_link_output = get_output(
                    Kathara.get_instance().exec(
                        machine_name=device_name,
                        lab_hash=lab.hash,
                        command="ip -d -j link",
                    )
                )
                bridge_vlan_output = get_output(
                    Kathara.get_instance().exec(
                        machine_name=device_name,
                        lab_hash=lab.hash,
                        command="bridge -j vlan",
                    )
                )
            except MachineNotRunningError as e:
                return [CheckResult(self.description, False, str(e))]

            actual_interfaces = json.loads(ip_link_output)

            actual_vlans = json.loads(bridge_vlan_output)

            for bridge_conf in bridges_configuration:
                expected_bridge_interfaces = list(bridge_conf["interfaces"].keys())
                vxlan_interfaces = []
                vxlan_interfaces_names = []
                if "vxlan" in bridge_conf:
                    for vni in bridge_conf["vxlan"]:
                        try:
                            interface = get_inteface_by_vni(vni, actual_interfaces)
                        except IndexError:
                            continue
                        vxlan_interfaces_names.append(interface["ifname"])
                        vxlan_interfaces.append(interface)
                    expected_bridge_interfaces.extend(vxlan_interfaces_names)

                check_result, masters = self.check_bridge_interfaces(
                    device_name, expected_bridge_interfaces, actual_interfaces
                )
                results.append(check_result)

                if check_result.passed:
                    check_result = self.check_vlan_filtering(
                        device_name, get_interface_by_name(masters.pop(), actual_interfaces)
                    )
                    results.append(check_result)

                    for interface_name, interface_conf in bridge_conf["interfaces"].items():
                        description = f"Getting VLAN info for `{interface_name}` on `{device_name}`"
                        actual_interface_vlans = None
                        try:
                            actual_interface_vlans = get_interface_by_name(interface_name, actual_vlans)
                            check_result = CheckResult(description, True, "OK")
                            results.append(check_result)
                        except IndexError:
                            check_result = CheckResult(
                                description,
                                False,
                                f"No VLAN found for for `{interface_name}` on `{device_name}`",
                            )
                            results.append(check_result)

                        if actual_interface_vlans:
                            if "vlan_tags" in interface_conf:
                                check_result = self.check_vlan_tags(
                                    device_name, interface_name, interface_conf, actual_interface_vlans
                                )
                                results.append(check_result)

                            if "pvid" in interface_conf:
                                check_result = self.check_vlan_pvid(
                                    device_name,
                                    interface_name,
                                    interface_conf["pvid"],
                                    actual_interface_vlans,
                                )
                                results.append(check_result)

                    if "vxlan" in bridge_conf:
                        for vni, vni_info in bridge_conf["vxlan"].items():
                            check_result = self.check_vxlan_pvid(
                                device_name, vni, vni_info["pvid"], actual_interfaces, actual_vlans
                            )
                            results.append(check_result)
        return results



from Kathara.manager.Kathara import Kathara
from Kathara.model.Lab import Lab

from ...AbstractCheck import AbstractCheck
from ...CheckResult import CheckResult
from ....utils import get_output


class DNSRecordCheck(AbstractCheck):

    def run(
            self,
            records: dict[str, dict[str, list[str]]],
            machines_with_dns: list[str],
            lab: Lab,
    ) -> list[CheckResult]:
        results = []
        kathara_manager: Kathara = Kathara.get_instance()

        for recordtype, recordvalue in records.items():
            for record, addresses in recordvalue.items():
                for client in machines_with_dns:
                    exec_output_gen = kathara_manager.exec(
                        machine_name=client,
                        command=f"dig +short {recordtype} {record}",
                        lab_hash=lab.hash,
                    )
                    ip = get_output(exec_output_gen).strip()
                    if ip in addresses:
                        check_result = CheckResult("Checking correctness of DNS records", True, "OK")
                    else:
                        check_result = CheckResult(
                            "Checking correctness of DNS records",
                            False,
                            f"{client} resolve {recordtype} {record} with IP {ip} instead of {addresses}",
                        )
                        check_result
                    results.append(check_result)
        return results



import logging
from abc import ABC


class AbstractCheck(ABC):

    def __init__(self, description: str = None):
        self.description: str = description
        self.logger = logging.getLogger("kathara-lab-checker")



import json

from Kathara.exceptions import MachineNotRunningError
from Kathara.manager.Kathara import Kathara
from Kathara.model.Lab import Lab

from ...AbstractCheck import AbstractCheck
from ...CheckResult import CheckResult
from ....utils import get_output


class VTEPCheck(AbstractCheck):
    def check(self, device_name: str, vni: str, vtep_ip: str, lab: Lab) -> CheckResult:
        kathara_manager: Kathara = Kathara.get_instance()
        try:
            exec_output_gen = kathara_manager.exec(
                machine_name=device_name, command="ip -d -j link show type vxlan", lab_hash=lab.hash
            )
        except MachineNotRunningError as e:
            return CheckResult(self.description, False, str(e))

        output = get_output(exec_output_gen)

        if output.startswith("ERROR:") or "exec failed" in output:
            return CheckResult(self.description, False, output)
        output = json.loads(output)

        for route in output:
            if route['linkinfo']['info_data']['id'] == int(vni):
                if route['linkinfo']['info_data']['local'] == vtep_ip:
                    return CheckResult(self.description, True, "OK")
                else:
                    reason = (f"VNI `{vni}` configured on device `{device_name}` with wrong "
                              f"VTEP IP {route['linkinfo']['info_data']['local']} (instead of {vtep_ip})")
                    return CheckResult(self.description, False, reason)
        return CheckResult(self.description, False, f"VNI `{vni}` not configured on device `{device_name}`")

    def run(self, device_to_vnis_info: dict[str, dict], lab: Lab) -> list[CheckResult]:
        results = []
        for device_name, vnis_info in device_to_vnis_info.items():
            self.logger.info(f"Checking {device_name} VTEP configuration...")
            vnis = vnis_info['vnis']
            vtep_ip = vnis_info['ip']
            for vni in vnis:
                self.description = f"Checking that `{device_name}` VTEP has vni `{vni}` with VTEP IP `{vtep_ip}`"
                check_result = self.check(device_name, vni, vtep_ip, lab)
                results.append(check_result)
        return results



import ipaddress
import json
from typing import Union, Any

from Kathara.exceptions import MachineNotRunningError
from Kathara.manager.Kathara import Kathara
from Kathara.model.Lab import Lab

from .AbstractCheck import AbstractCheck
from .CheckResult import CheckResult


def load_routes_from_expected(expected_routes: list) -> dict[str, set]:
    routes = {}
    for route in expected_routes:
        if type(route) is list:
            routes[route[0]] = set(route[1])
        else:
            routes[route] = set()
    return routes


def get_kernel_routes(device_name: str, lab: Lab) -> list[dict[str, Any]]:
    kathara_manager = Kathara.get_instance()
    try:
        stdout, _, _ = kathara_manager.exec(machine_name=device_name, lab_hash=lab.hash, command="ip -j route",
                                            stream=False)
        stdout = stdout.decode("utf-8").strip()
    except MachineNotRunningError:
        return []
    return json.loads(stdout)


def get_nexthops(device_name: str, lab: Lab) -> list[dict[str, Any]]:
    kathara_manager = Kathara.get_instance()

    try:
        stdout, _, _ = kathara_manager.exec(machine_name=device_name, lab_hash=lab.hash, command="ip -j nexthop",
                                            stream=False)
        stdout = stdout.decode("utf-8").strip()
    except MachineNotRunningError:
        return []

    return json.loads(stdout)


def load_routes_from_device(device_name: str, lab: Lab) -> dict[str, set]:
    ip_route_output = get_kernel_routes(device_name, lab)
    routes = {}
    kernel_nexthops = None

    for route in ip_route_output:

        dst = route["dst"]
        if dst == "default":
            dst = "0.0.0.0/0"
        nexthops = None
        if "scope" in route and route["scope"] == "link":
            nexthops = [("d.c.", route["dev"])]
        elif "nexthops" in route:
            nexthops = list(map(lambda x: x["dev"], route["nexthops"]))
        elif "gateway" in route:
            nexthops = [(route["gateway"], route["dev"])]
        elif "via" in route:
            nexthops = [(route["via"]["host"], route["dev"])]
        elif "nhid" in route:
            # Lazy load nexthops
            kernel_nexthops = get_nexthops(device_name, lab) if kernel_nexthops is None else kernel_nexthops

            current_nexthop = [obj for obj in kernel_nexthops if obj["id"] == route["nhid"]][0]
            if "gateway" in current_nexthop:
                nexthops = [(current_nexthop["gateway"], current_nexthop["dev"])]
            elif "group" in current_nexthop:
                nexthops = [
                    (obj["gateway"], obj["dev"])
                    for obj in kernel_nexthops
                    if obj["id"] in (nhid["id"] for nhid in current_nexthop["group"])
                ]
            else:
                raise Exception("Strange nexthop: ", current_nexthop)
        routes[dst] = set(nexthops)
    return routes


def is_valid_ip(ip_str):
    try:
        ipaddress.ip_address(ip_str)
        return True
    except ValueError:
        return False


class KernelRouteCheck(AbstractCheck):
    def check(self, device_name: str, expected_routing_table: list, lab: Lab) -> list[CheckResult]:
        self.description = f"Checking the routing table of {device_name}"
        actual_routing_table = dict(
            filter(
                lambda item: not any("d.c." in elem for elem in item[1]),
                load_routes_from_device(device_name, lab).items(),
            )
        )
        expected_routing_table = load_routes_from_expected(expected_routing_table)

        results = []

        if len(expected_routing_table) != len(actual_routing_table):
            check_result = CheckResult(
                self.description,
                False,
                f"The routing table of {device_name} have the wrong number of routes: {len(actual_routing_table)}, expected: {len(expected_routing_table)}",
            )
            results.append(check_result)
            return results

        for dst, nexthops in expected_routing_table.items():
            if not dst in actual_routing_table:
                check_result = CheckResult(
                    self.description, False, f"The routing table of {device_name} is missing route {dst}"
                )
                results.append(check_result)
                continue
            if nexthops:
                actual_nh = actual_routing_table[dst]
                if len(nexthops) != len(actual_nh):
                    check_result = CheckResult(
                        self.description,
                        False,
                        f"The routing table of {device_name} about route {dst} have the wrong number of next-hops: {len(actual_nh)}, expected: {len(nexthops)}",
                    )
                    results.append(check_result)
                    continue
                for nh in nexthops:
                    valid_ip = is_valid_ip(nh)
                    if (valid_ip and not any(item[0] == nh for item in actual_nh)) or (
                            not valid_ip and not any(item[1] == nh for item in actual_nh)
                    ):
                        check_result = CheckResult(
                            self.description,
                            False,
                            f"The routing table of {device_name} about route {dst} does not contain next-hop: {nh}, actual: {actual_nh}",
                        )
                        results.append(check_result)

        if not results:
            check_result = CheckResult(self.description, True, f"OK")
            results.append(check_result)

        return results

    def run(self, devices_to_routes: dict[str, list[Union[str, list[str]]]], lab: Lab) -> list[CheckResult]:
        results = []
        for device_name, expected_routes in devices_to_routes.items():
            self.logger.info(f"Checking kernel routes for `{device_name}`...")
            try:
                check_result = self.check(device_name, expected_routes, lab)
                results = results + check_result
            except MachineNotRunningError:
                self.logger.warning(f"`{device_name}` is not running. Skipping checks...")
        return results



import json

from Kathara.exceptions import MachineNotRunningError
from Kathara.manager.Kathara import Kathara
from Kathara.model.Lab import Lab

from ...AbstractCheck import AbstractCheck
from ...CheckResult import CheckResult
from ....utils import get_output


class EVPNSessionCheck(AbstractCheck):
    def check(self, device_name: str, neighbor: str, lab: Lab) -> CheckResult:
        kathara_manager: Kathara = Kathara.get_instance()
        try:
            exec_output_gen = kathara_manager.exec(
                machine_name=device_name, command="vtysh -e 'show bgp summary json'", lab_hash=lab.hash
            )
        except MachineNotRunningError as e:
            return CheckResult(self.description, False, str(e))

        output = get_output(exec_output_gen)

        if output.startswith("ERROR:") or "exec failed" in output:
            return CheckResult(self.description, False, output)
        output = json.loads(output)
        if "l2VpnEvpn" in output:
            try:
                for peer_name, peer in output["l2VpnEvpn"]["peers"].items():
                    if neighbor == peer_name:
                        if peer["state"] == "Established":
                            return CheckResult(self.description, True, "OK")
                        else:
                            return CheckResult(
                                self.description,
                                False,
                                f"The session is configured but is in the {peer['state']} state",
                            )
            except KeyError:
                pass
            reason = f"The evpn session between {device_name} and {neighbor} is not up."

            return CheckResult(self.description, False, reason)

        else:
            return CheckResult(
                self.description, False, f"`l2VpnEvpn` address family not active for bgp on {device_name}"
            )

    def run(self, device_to_neighbours: dict[str, list[str]], lab: Lab) -> list[CheckResult]:
        results = []
        for device_name, neighbors in device_to_neighbours.items():
            self.logger.info(f"Checking that {device_name} has `address-family l2vpn evpn` activated...")
            for neighbor in neighbors:
                self.description = f"{device_name} has bgp peer {neighbor}"
                check_result = self.check(device_name, neighbor, lab)
                results.append(check_result)
        return results



from Kathara.exceptions import MachineNotFoundError
from Kathara.manager.Kathara import Kathara
from Kathara.model.Lab import Lab

from ..utils import get_output
from .AbstractCheck import AbstractCheck
from .CheckResult import CheckResult


class DaemonCheck(AbstractCheck):

    def check(self, device_name: str, daemon: str, lab: Lab) -> CheckResult:
        kathara_manager: Kathara = Kathara.get_instance()

        if daemon.startswith("!"):
            daemon = daemon[1:]
            self.description = f"Checking that {daemon} is not running on device `{device_name}`"
            invert = True
        else:
            self.description = f"Checking that {daemon} is running on device `{device_name}`"
            invert = False

        try:
            device = lab.get_machine(device_name)
            output = get_output(
                kathara_manager.exec(machine_name=device.name, lab_hash=lab.hash, command=f"pgrep {daemon}")
            )
            if (output != "") ^ invert:
                return CheckResult(self.description, True, "OK")
            else:
                reason = f"Daemon {daemon} is {'' if invert else 'not '}running on device `{device_name}`"
                return CheckResult(self.description, False, reason)
        except MachineNotFoundError as e:
            return CheckResult(self.description, False, str(e))

    def run(self, devices_to_daemons: dict[str, list[str]], lab: Lab) -> list[CheckResult]:
        results = []
        for device_name, daemons in devices_to_daemons.items():
            self.logger.info(f"Checking if daemons are running on `{device_name}`...")
            for daemon_name in daemons:
                check_result = self.check(device_name, daemon_name, lab)
                results.append(check_result)
        return results



import ipaddress

from Kathara.exceptions import MachineNotRunningError
from Kathara.model.Lab import Lab

from ..utils import get_interfaces_addresses
from .AbstractCheck import AbstractCheck
from .CheckResult import CheckResult


class InterfaceIPCheck(AbstractCheck):

    def check(self, device_name: str, interface_number: int, ip: str, dumped_iface: dict) -> CheckResult:
        interface_name = f"eth{interface_number}" if interface_number.isnumeric() else interface_number
        self.description = f"Verifying the IP address ({ip}) assigned to {interface_name} of {device_name}"

        try:
            iface_info = next(filter(lambda x: x["ifname"] == f"{interface_name}", dumped_iface))
        except StopIteration:
            return CheckResult(self.description, False, f"Interface `{interface_name}` not found on `{device_name}`")

        ip_address = ipaddress.ip_interface(ip)

        prefix_len = int(ip_address.with_prefixlen.split("/")[1])

        assigned_ips = []
        if iface_info and "addr_info" in iface_info:
            for addr_info in iface_info["addr_info"]:
                assigned_ips.append(f"{addr_info['local']}/{addr_info['prefixlen']}")
                if addr_info["local"] == str(ip_address.ip):
                    if addr_info["prefixlen"] == prefix_len:
                        return CheckResult(self.description, True, "OK")
                    else:
                        reason = f"The IP address has a wrong netmask ({addr_info['prefixlen']})"
                        return CheckResult(self.description, False, reason)

        reason = (
            f"The interface `{iface_info['ifname']}` of `{device_name}` "
            f"has the following IP addresses: {assigned_ips}`."
        )
        return CheckResult(self.description, False, reason)

    def run(self, ip_mapping: dict, lab: Lab) -> list[CheckResult]:
        results = []
        for device_name, iface_to_ips in ip_mapping.items():
            self.logger.info(f"Checking IPs for `{device_name}`...")
            try:
                dumped_iface = get_interfaces_addresses(device_name, lab)
                for interface_number, ip in iface_to_ips.items():
                    check_result = self.check(device_name, interface_number, ip, dumped_iface)
                    results.append(check_result)
            except MachineNotRunningError:
                self.logger.warning(f"`{device_name}` is not running. Skipping checks...")
        return results



import re

from Kathara.manager.Kathara import Kathara
from Kathara.model.Lab import Lab

from ..AbstractCheck import AbstractCheck
from ..CheckResult import CheckResult
from ...utils import get_output


class AnnouncedNetworkCheck(AbstractCheck):
    def check(self, device_name: str, protocol: str, network: str, lab: Lab) -> CheckResult:
        self.description = f"Check {protocol} network ({network}) for {device_name}"

        kathara_manager: Kathara = Kathara.get_instance()

        try:
            exec_output_gen = kathara_manager.exec(
                machine_name=device_name, command=f"vtysh -e 'show running-config {protocol}'", lab_hash=lab.hash
            )
        except Exception as e:
            return CheckResult(self.description, False, str(e))

        output = list(filter(lambda x: "network" in x, get_output(exec_output_gen).split("\n")))
        for line in output:
            if re.search(rf"\s*network\s*{network}", line):
                return CheckResult(self.description, True, "OK")
        reason = f"Network {network} is not announced in {protocol}."
        return CheckResult(self.description, False, reason)

    def run(self, protocol: str, devices_to_networks: dict[str, list[str]], lab: Lab) -> list[CheckResult]:
        results = []
        for device_name, networks in devices_to_networks.items():
            self.logger.info(f"Checking {device_name} BGP announces...")
            for network in networks:
                check_result = self.check(device_name, protocol, network, lab)
                results.append(check_result)
        return results



import re

from Kathara.manager.Kathara import Kathara
from Kathara.model.Lab import Lab

from ..AbstractCheck import AbstractCheck
from ..CheckResult import CheckResult
from ...utils import get_output


class ProtocolRedistributionCheck(AbstractCheck):

    def check(self, device_name: str, protocol_to_check: str, injected_protocol: str, lab: Lab) -> CheckResult:
        kathara_manager: Kathara = Kathara.get_instance()

        if injected_protocol.startswith("!"):
            injected_protocol = injected_protocol[1:]
            self.description = f"Checking that {injected_protocol} routes are not redistributed to {protocol_to_check} on device `{device_name}`"
            invert = True
        else:
            self.description = f"Checking that {injected_protocol} routes are redistributed to {protocol_to_check} on device `{device_name}`"
            invert = False

        try:
            exec_output_gen = kathara_manager.exec(
                machine_name=device_name,
                command=f"vtysh -e 'show running-config {protocol_to_check}'",
                lab_hash=lab.hash,
            )
        except Exception as e:
            return CheckResult(self.description, False, str(e))

        output = get_output(exec_output_gen).split("\n")
        found = False
        for line in output:
            if re.search(rf"^\s*redistribute\s*{injected_protocol}$", line):
                found = True
                break
        if found ^ invert:
            return CheckResult(self.description, True, "OK")
        else:
            reason = f"{injected_protocol} routes are {'' if invert else 'not '}injected into `{protocol_to_check}` on `{device_name}`."
        return CheckResult(self.description, False, reason)

    def run(self, protocol, devices_to_redistributed: dict[str, list[str]], lab: Lab) -> list[CheckResult]:
        results = []
        for device_name, injected_protocols in devices_to_redistributed.items():
            for injected_protocol in injected_protocols:
                check_result = self.check(device_name, protocol, injected_protocol, lab)
                results.append(check_result)
        return results



from Kathara.exceptions import MachineNotFoundError
from Kathara.model.Lab import Lab

from .AbstractCheck import AbstractCheck
from .CheckResult import CheckResult


class IPv6EnabledCheck(AbstractCheck):
    def check(self, device_name: str, lab: Lab) -> CheckResult:
        self.description = f"Checking the IPv6 is enabled on `{device_name}`"

        try:
            device = lab.get_machine(device_name)
            if "ipv6" in device.meta and device.is_ipv6_enabled():
                return CheckResult(self.description, True, "OK")
            else:
                return CheckResult(self.description, False, f"IPv6 not enabled on `{device_name}`")
        except MachineNotFoundError as e:
            return CheckResult(self.description, False, str(e))

    def run(self, ipv6_devices: list[str], lab: Lab) -> list[CheckResult]:
        results = []
        for device_name in ipv6_devices:
            check_result = self.check(device_name, lab)
            results.append(check_result)
        return results



from Kathara.model.Lab import Lab

from .AbstractCheck import AbstractCheck
from .CheckResult import CheckResult


class StartupExistenceCheck(AbstractCheck):

    def check(self, device_name: str, lab: Lab) -> CheckResult:
        self.description = f"Check existence of `{device_name}.startup` file"

        if lab.fs.exists(device_name + ".startup"):
            return CheckResult(self.description, True, "OK")
        else:
            return CheckResult(self.description, False, f"{device_name}.startup file not found")

    def run(self, machines_to_check: list[str], lab: Lab) -> list[CheckResult]:
        results = []
        for device_name in machines_to_check:
            check_result = self.check(device_name, lab)
            results.append(check_result)
        return results



import json
from typing import Optional

from Kathara.manager.Kathara import Kathara
from Kathara.model.Lab import Lab


def red(s):
    return f"\033[91m {s}\033[00m"


def green(s):
    return f"\033[92m {s}\033[00m"


def yellow(s):
    return f"\033[93m {s}\033[00m"


def get_output(exec_output):
    output = ""
    try:
        while True:
            (stdout, stderr) = next(exec_output)
            stdout = stdout.decode("utf-8") if stdout else ""
            stderr = stderr.decode("utf-8") if stderr else ""

            if stdout:
                output += stdout
            if stderr:
                output += f"ERROR: {stderr}"
    except StopIteration:
        pass
    return output


def get_interfaces_addresses(device_name: str, lab: Lab) -> dict:
    kathara_manager = Kathara.get_instance()

    exec_output_gen = kathara_manager.exec(
        machine_name=device_name,
        command=f"ip -j address",
        lab_hash=lab.hash,
    )

    return json.loads(get_output(exec_output_gen))


def find_device_name_from_ip(ip_mapping: dict[str, dict[str, str]], ip_search: str) -> Optional[str]:
    for device, ip_addresses in ip_mapping.items():
        for _, ip in ip_addresses.items():
            # Check if the base IP matches (ignoring the CIDR notation)
            if ip.split("/")[0] == ip_search:
                return device
    raise Exception("Something is missing/wrong in the ip_mapping configuration!")


def find_lines_with_string(file_content, search_string):
    """
    Returns lines from the provided multi-line string that contain the search string.

    :param file_content: A string representing the content of a file (multi-line string).
    :param search_string: A string to search for in each line of the file content.
    :return: A list of lines that contain the search string.
    """
    # Splitting the string into lines
    lines = file_content.split("\n")

    # Filtering lines that contain the search string
    matching_lines = [line for line in lines if search_string in line]

    return matching_lines


def reverse_dictionary(dictionary: dict):
    reversed_dict = {}
    for k, values in dictionary.items():
        for v in values:
            reversed_dict[v] = reversed_dict.get(v, []) + [k]
    return reversed_dict



import logging

from ..utils import green, red


class CheckResult:

    def __init__(self, description: str, passed: bool, reason: str) -> None:
        self.description: str = description
        self.passed: bool = passed
        self.reason: str = reason
        logging.getLogger("kathara-lab-checker").info(self)

    def __str__(self) -> str:
        return f"{self.description}: {green(self.reason) if self.passed else red(self.reason)}"



import re

import jc
import os
from Kathara.exceptions import MachineNotRunningError
from Kathara.manager.Kathara import Kathara
from Kathara.model.Lab import Lab

from ...AbstractCheck import AbstractCheck
from ...CheckResult import CheckResult
from ....utils import get_output, find_lines_with_string, find_device_name_from_ip


class DNSAuthorityCheck(AbstractCheck):
    def check(self, domain: str, authority_ip: str, device_name: str, device_ip: str, lab: Lab) -> CheckResult:
        self.description = f"Checking on `{device_name}` that `{authority_ip}` is the authority for domain `{domain}`"
        kathara_manager: Kathara = Kathara.get_instance()
        try:
            exec_output_gen = kathara_manager.exec(
                machine_name=device_name, command=f"dig NS {domain} @{device_ip}", lab_hash=lab.hash
            )
        except MachineNotRunningError as e:
            return CheckResult(self.description, False, str(e))

        output = get_output(exec_output_gen)
        if output.startswith("ERROR:"):
            return CheckResult(self.description, False, output)

        result = jc.parse("dig", output)
        if result:
            result = result.pop()
            if result["status"] == "NOERROR" and "answer" in result:
                root_servers = list(map(lambda x: x["data"].split(" ")[0], result["answer"]))
                authority_ips = []
                for root_server in root_servers:
                    exec_output_gen = kathara_manager.exec(
                        machine_name=device_name,
                        command=f"dig +short {root_server} @{device_ip}",
                        lab_hash=lab.hash,
                    )
                    ip = get_output(exec_output_gen).strip()
                    if authority_ip == ip:
                        return CheckResult(self.description, True, "OK")
                    else:
                        authority_ips.append(ip)
                reason = f"The dns authorities for domain `{domain}` have the following IPs {authority_ips}"
                return CheckResult(self.description, False, reason)
            else:
                reason = (
                    f"named on {device_name} is running but answered "
                    f"with {result['status']} when quering for {domain}"
                )
                return CheckResult(self.description, False, reason)
        else:
            if lab.fs.exists(f"{device_name}.startup"):
                with lab.fs.open(f"{device_name}.startup", "r") as startup_file:
                    lines = startup_file.readlines()

                for line in lines:
                    line = line.strip()
                    if re.search(rf"^\s*systemctl\s*start\s*named\s*$", line):
                        exec_output_gen = kathara_manager.exec(
                            machine_name=device_name,
                            command=f"named -d 5 -g",
                            lab_hash=lab.hash,
                        )

                        output = get_output(exec_output_gen)
                        date_pattern = (
                            r"\d{2}-[Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec]{3}-\d{4} \d{2}:\d{2}:\d{2}\.\d{3}"
                        )

                        reason_list = find_lines_with_string(output, "could not")
                        reason_list.extend(find_lines_with_string(output, "/etc/bind/named.conf"))
                        reason_list_no_dates = [re.sub(date_pattern, "", line) for line in reason_list]
                        reason = "\n".join(reason_list_no_dates)

                        return CheckResult(self.description, False, "Configuration Error:\n" + reason)

                reason = f"named not started in `{device_name}`.startup`"
                return CheckResult(self.description, False, reason)
            else:
                reason = f"There is no `.startup` file for device `{device_name}`"
                return CheckResult(self.description, False, reason)

    def run(
        self,
        zone_to_authoritative_ips: dict[str, list[str]],
        local_nameservers: list[str],
        ip_mapping: dict[str, dict[str, str]],
        lab: Lab,
    ) -> list[CheckResult]:
        results = []
        for domain, name_servers in zone_to_authoritative_ips.items():
            self.logger.info(f"Checking authority ip for domain `{domain}`")
            for ns in name_servers:
                check_result = self.check(domain, ns, find_device_name_from_ip(ip_mapping, ns), ns, lab)
                results.append(check_result)

                if domain == ".":
                    self.logger.info(
                        f"Checking if all the named servers can correctly resolve {ns} as the root nameserver..."
                    )
                    for generic_ns_ip in name_servers:
                        check_result = self.check(
                            domain,
                            ns,
                            find_device_name_from_ip(ip_mapping, generic_ns_ip),
                            generic_ns_ip,
                            lab,
                        )
                        results.append(check_result)

                    for local_ns in local_nameservers:
                        check_result = self.check(
                            domain, ns, find_device_name_from_ip(ip_mapping, local_ns), local_ns, lab
                        )
                        results.append(check_result)
        return results



from Kathara.exceptions import MachineNotFoundError
from Kathara.model.Lab import Lab

from .AbstractCheck import AbstractCheck
from .CheckResult import CheckResult


class DeviceExistenceCheck(AbstractCheck):

    def check(self, device_name: str, lab: Lab) -> CheckResult:
        self.description = f'Check existence of `{device_name}`'
        try:
            lab.get_machine(device_name)
        except MachineNotFoundError as e:
            return CheckResult(self.description, False, str(e))

        return CheckResult(self.description, True, "OK")

    def run(self, template_machines: list[str], lab: Lab) -> list[CheckResult]:
        results = []
        for device_name in template_machines:
            check_result = self.check(device_name, lab)
            results.append(check_result)
        return results



import os

from openpyxl import Workbook
from openpyxl.styles import Alignment
from openpyxl.worksheet.worksheet import Worksheet


def write_final_results_to_excel(test_collector: "TestCollectorPackage.TestCollector", path: str):
    # Create a new Excel workbook
    workbook = Workbook()

    # Select the active sheet
    sheet = workbook.active

    sheet["A1"] = "Student Name"
    sheet["B1"] = "Tests Passed"
    sheet["C1"] = "Tests Failed"
    sheet["D1"] = "Tests Total Number"
    sheet["E1"] = "Problems"

    for index, (test_name, test_results) in enumerate(test_collector.tests.items()):
        failed_tests = test_collector.get_failed(test_name)
        passed_tests = test_collector.get_passed(test_name)
        sheet["A" + str(index + 2)] = test_name
        sheet["B" + str(index + 2)] = len(passed_tests)
        sheet["C" + str(index + 2)] = len(failed_tests)
        sheet["D" + str(index + 2)] = len(test_results)

        if failed_tests:
            failed_string = ""
            for idx, failed in enumerate(failed_tests):
                failed_string += f"{(idx + 1)}: {failed.description}: {failed.reason}\n"
            if len(failed_string) >= 32767:
                raise Exception("ERROR: Excel cell too big")
            sheet["E" + str(index + 2)] = failed_string
            sheet["E" + str(index + 2)].alignment = Alignment(wrapText=True)
        else:
            sheet["E" + str(index + 2)] = "None"

    excel_file = os.path.join(path, "results.xlsx")
    workbook.save(excel_file)


def _write_sheet_row(sheet: Worksheet, column: int, description: str, passed: str, reason: str) -> None:
    sheet["A" + str(column + 2)] = description
    sheet["B" + str(column + 2)] = passed
    sheet["C" + str(column + 2)] = reason


def write_result_to_excel(check_results: list["CheckResultPackage.CheckResult"], path: str):
    # Create a new Excel workbook
    workbook: Workbook = Workbook()

    workbook.create_sheet("Summary", 0)
    sheet_summary = workbook.get_sheet_by_name("Summary")
    sheet_summary["A1"] = "Total Tests"
    sheet_summary["B1"] = "Passed Tests"
    sheet_summary["C1"] = "Failed"

    _write_sheet_row(
        sheet_summary,
        0,
        str(len(check_results)),
        str(len(list(filter(lambda x: x.passed, check_results)))),
        str(len(list(filter(lambda x: not x.passed, check_results)))),
    )

    # Select the active sheet
    workbook.create_sheet("All", 1)
    sheet_all = workbook.get_sheet_by_name("All")
    sheet_all["A1"] = "Tests Description"
    sheet_all["B1"] = "Passed"
    sheet_all["C1"] = "Reason"

    workbook.create_sheet("Failed", 2)
    sheet_failed = workbook.get_sheet_by_name("Failed")
    sheet_failed["A1"] = "Tests Description"
    sheet_failed["B1"] = "Passed"
    sheet_failed["C1"] = "Reason"

    failed_index = 0
    for index, check_result in enumerate(check_results):
        if not check_result.passed:
            _write_sheet_row(
                sheet_failed,
                failed_index,
                check_result.description,
                str(check_result.passed),
                check_result.reason,
            )
            failed_index += 1
        _write_sheet_row(sheet_all, index, check_result.description, str(check_result.passed), check_result.reason)
    workbook.save(os.path.join(path, f"{os.path.basename(path)}_result.xlsx"))



import re

from Kathara.manager.Kathara import Kathara
from Kathara.model.Lab import Lab

from ...AbstractCheck import AbstractCheck
from ...CheckResult import CheckResult
from ....utils import get_output


class LocalNSCheck(AbstractCheck):

    def check(self, local_ns_ip: str, device_name: str, lab: Lab) -> CheckResult:
        kathara_manager: Kathara = Kathara.get_instance()

        self.description = f"Checking that `{local_ns_ip}` is the local name server for device `{device_name}`"

        exec_output_gen = kathara_manager.exec(
            machine_name=device_name, command=f"cat /etc/resolv.conf", lab_hash=lab.hash
        )
        output = get_output(exec_output_gen)
        if output.startswith("ERROR:"):
            return CheckResult(self.description, False, output)

        lines = output.splitlines()
        if not lines:
            reason = f"`resolv.conf` file not found for device `{device_name}`"
            return CheckResult(self.description, False, reason)
        actual_ips = []
        for line in lines:
            match = re.search(rf"^nameserver (.*)$", line)
            if match:
                actual_ns_ip = match.group(1)
                if actual_ns_ip == local_ns_ip:
                    return CheckResult(self.description, True, "OK")
                actual_ips.append(actual_ns_ip)

        reason = (f"There is no local name server for device `{device_name}` with IP `{local_ns_ip}`. "
                  f"Actual nameservers: {actual_ips}")
        return CheckResult(self.description, False, reason)
    def run(self, local_nameservers_to_devices: dict[str, list[str]], lab: Lab) -> list[CheckResult]:
        results = []
        for local_ns, managed_devices in local_nameservers_to_devices.items():
            for device_name in managed_devices:
                check_result = self.check(local_ns, device_name, lab)
                results.append(check_result)
        return results



import re

from Kathara.manager.Kathara import Kathara
from Kathara.model.Lab import Lab

from ...AbstractCheck import AbstractCheck
from ...CheckResult import CheckResult
from ....utils import get_output


class AnnouncedVNICheck(AbstractCheck):
    def check(self, device_name: str, lab: Lab, invert: bool = False) -> CheckResult:
        if not invert:
            self.description = f"Check that {device_name} announces all VNIs"
        else:
            self.description = f"Check that {device_name} not announces any VNIs"

        kathara_manager: Kathara = Kathara.get_instance()

        try:
            exec_output_gen = kathara_manager.exec(
                machine_name=device_name, command=f"vtysh -e 'show running-config bgpd'", lab_hash=lab.hash
            )
        except Exception as e:
            return CheckResult(self.description, False, str(e))

        output = get_output(exec_output_gen).splitlines()
        for line in output:
            if re.search(rf"\s*advertise-all-vni\s*", line):
                if invert:
                    return CheckResult(self.description, False,
                                       f"`advertise-all-vni` found in `{device_name}` bgpd configuration")
                else:
                    return CheckResult(self.description, True, "OK")

        if invert:
            return CheckResult(self.description, True, "OK")
        else:
            return CheckResult(self.description, False,
                               f"`advertise-all-vni` not found in `{device_name}` bgpd configuration")

    def run(self, device_to_vnis_info: dict[str, dict], evpn_devices: list[str], lab: Lab) -> list[CheckResult]:
        results = []
        for device_name in device_to_vnis_info.keys():
            check_result = self.check(device_name, lab)
            results.append(check_result)

        not_advertise = set(evpn_devices).difference(set(device_to_vnis_info.keys()))
        for device_name in not_advertise:
            check_result = self.check(device_name, lab, invert=True)
            results.append(check_result)

        return results



from Kathara.exceptions import MachineNotFoundError
from Kathara.model.Lab import Lab

from .AbstractCheck import AbstractCheck
from .CheckResult import CheckResult


class SysctlCheck(AbstractCheck):
    def check(self, device_name: str, sysctl_to_check: str, lab: Lab) -> CheckResult:
        self.description = f"Checking that sysctl `{sysctl_to_check}` is set on `{device_name}`"

        try:
            device = lab.get_machine(device_name)
            split = sysctl_to_check.split("=")
            sysctl_to_check, value_to_check = split[0], int(split[1])

            device_sysctls = device.get_sysctls()

            if sysctl_to_check in device_sysctls:
                if value_to_check == device_sysctls[sysctl_to_check]:
                    return CheckResult(self.description, True, "OK")
                else:
                    return CheckResult(self.description, False,
                                       f"Sysctl `{sysctl_to_check}` set on `{device_name}` with wrong "
                                       f"value `{device_sysctls[sysctl_to_check]}` (instead of `{value_to_check}`)")
            else:
                return CheckResult(self.description, False, f"Sysctl `{sysctl_to_check}` not set on `{device_name}`")
        except MachineNotFoundError as e:
            return CheckResult(self.description, False, str(e))

    def run(self, devices_sysctls: dict[str, list[str]], lab: Lab) -> list[CheckResult]:
        results = []
        for device_name, sysctls in devices_sysctls.items():
            for sysctl in sysctls:
                check_result = self.check(device_name, sysctl, lab)
                results.append(check_result)
        return results



import logging
import re

from Kathara.exceptions import MachineNotFoundError
from Kathara.manager.Kathara import Kathara
from Kathara.model.Lab import Lab

from .AbstractCheck import AbstractCheck
from .CheckResult import CheckResult


class CustomCommandCheck(AbstractCheck):

    def check(self, device_name: str, command_entry: dict[str, str | int], lab: Lab) -> list[CheckResult]:
        kathara_manager: Kathara = Kathara.get_instance()

        results = []
        try:
            device = lab.get_machine(device_name)
            stdout, stderr, exit_code = kathara_manager.exec(
                machine_name=device.name, lab_hash=lab.hash, command=command_entry["command"], stream=False
            )

            stdout = stdout.decode("utf-8").strip() if stdout else (stderr.decode("utf-8").strip() if stderr else "")

            if "exit_code" in command_entry:
                self.description = (
                    f"Checking the exit code of the command '{command_entry['command']}' on '{device_name}'"
                )
                if exit_code == command_entry["exit_code"]:
                    results.append(CheckResult(self.description, True, "OK"))
                else:
                    reason = (
                        f"The exit code of the command differs from the expected one."
                        f"\n Actual: {exit_code}\n Expected: {command_entry['exit_code']}"
                    )
                    results.append(CheckResult(self.description, False, reason))

            self.description = f"Checking the output of the command '{command_entry['command']}' on '{device_name}'"
            if "output" in command_entry:
                stdout = stdout.replace("\r\n", "\n").replace("\r", "\n")
                command_entry["output"] = command_entry["output"].replace("\r\n", "\n").replace("\r", "\n")

                if stdout == command_entry["output"].replace("\r\n", "\n").replace("\r", "\n"):
                    results.append(CheckResult(self.description, True, "OK"))
                else:
                    reason = (
                        f"The output of the command differs from the expected one."
                        f"\n Actual: {stdout}\n Expected: {command_entry['output']}"
                    )
                    results.append(CheckResult(self.description, False, reason))
            if "regex_match" in command_entry:

                if re.search(command_entry["regex_match"], stdout):
                    results.append(CheckResult(self.description, True, "OK"))
                else:
                    reason = (
                        f"The output of the command do not match the expected regex."
                        f"\n Actual: {stdout}\n Regex: {command_entry['regex_match']}"
                    )
                    results.append(CheckResult(self.description, False, reason))

        except MachineNotFoundError as e:
            results.append(CheckResult(self.description, False, str(e)))

        return results

    def run(self, devices_to_commands: dict[str, list[dict[str, str | int]]], lab: Lab) -> list[CheckResult]:
        results = []
        for device_name, command_entries in devices_to_commands.items():
            for command_entry in command_entries:
                check_result = self.check(device_name, command_entry, lab)
                results.extend(check_result)
        return results



import jc
from Kathara.manager.Kathara import Kathara
from Kathara.model.Lab import Lab

from ..utils import get_output
from .AbstractCheck import AbstractCheck
from .CheckResult import CheckResult


class ReachabilityCheck(AbstractCheck):

    def check(self, device_name: str, destination: str, lab: Lab) -> CheckResult:
        kathara_manager: Kathara = Kathara.get_instance()

        if destination.startswith("!"):
            destination = destination[1:]
            self.description = f"Verifying `{destination}` not reachable from device `{device_name}`"
            invert = True
        else:
            self.description = f"Verifying `{destination}` reachable from device `{device_name}`"
            invert = False

        try:
            exec_output_gen = kathara_manager.exec(
                machine_name=device_name,
                command=f"bash -c 'ping -q -n -c 1 {destination}'",
                lab_hash=lab.hash,
            )
        except Exception as e:
            return CheckResult(self.description, invert ^ False, str(e))

        output = get_output(exec_output_gen).replace("ERROR: ", "")

        try:
            parsed_output = jc.parse("ping", output, quiet=True)
            if int(parsed_output["packets_received"]) > 0:
                reason = f"`{device_name}` can reach `{destination}`." if invert else "OK"
                return CheckResult(self.description, invert ^ True, reason)
            else:
                reason = (
                    "OK" if invert else f"`{device_name}` does not receive any answer from `{destination}`."
                )
                return CheckResult(self.description, invert ^ False, reason)
        except Exception:
            return CheckResult(self.description, False, output.strip())

    def run(self, devices_to_destinations: dict[str, list[str]], lab: Lab) -> list[CheckResult]:
        results = []
        for device_name, destinations in devices_to_destinations.items():
            for destination in destinations:
                check_result = self.check(device_name, destination, lab)
                results.append(check_result)
        return results



class TestCollector:
    def __init__(self):
        self.tests = {}

    def add_check_result(self, lab_name: str, check_result: 'CheckResultPackage.CheckResult') -> None:
        if lab_name not in self.tests:
            self.tests[lab_name] = []
        self.tests[lab_name].append(check_result)

    def add_check_results(self, lab_name: str, check_results: list['CheckResultPackage.CheckResult']) -> None:
        for result in check_results:
            self.add_check_result(lab_name, result)

    def get_failed(self, test_name: str):
        return list(filter(lambda x: not x.passed, self.tests[test_name]))

    def get_passed(self, test_name: str):
        return list(filter(lambda x: x.passed, self.tests[test_name]))
